# Copyright 2023 The EASYDEL Author @erfanzar (Erfan Zare Chavoshi).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from easydel.infra.base_module import EasyDeLBaseConfig
from easydel.infra.factory import register_config, registry
from easydel.utils.helpers import get_logger

from ..auto import AutoEasyDeLConfig

logger = get_logger(__name__)


@register_config("llava")
class LlavaConfig(EasyDeLBaseConfig):
	r"""
	This is the configuration class to store the configuration of a [`LlavaForConditionalGeneration`]. It is used to instantiate an
	Llava model according to the specified arguments, defining the model architecture. Instantiating a configuration
	with the defaults will yield a similar configuration to that of the Llava-9B.

	e.g. [llava-hf/llava-9b](https://huggingface.co/llava-hf/llava-9b)

	Configuration objects inherit from [`PretrainedConfig`] and can be used to control the model outputs. Read the
	documentation from [`PretrainedConfig`] for more information.

	Args:
	    vision_config (`Union[AutoConfig, dict]`,  *optional*, defaults to `CLIPVisionConfig`):
	        The config object or dictionary of the vision backbone.
	    text_config (`Union[AutoConfig, dict]`, *optional*, defaults to `LlamaConfig`):
	        The config object or dictionary of the text backbone.
	    image_token_index (`int`, *optional*, defaults to 32000):
	        The image token index to encode the image prompt.
	    projector_hidden_act (`str`, *optional*, defaults to `"gelu"`):
	        The activation function used by the multimodal projector.
	    vision_feature_select_strategy (`str`, *optional*, defaults to `"default"`):
	        The feature selection strategy used to select the vision feature from the vision backbone.
	        Can be one of `"default"` or `"full"`.
	    vision_feature_layer (`Union[int, List[int]]`, *optional*, defaults to -2):
	        The index of the layer to select the vision feature. If multiple indices are provided,
	        the vision feature of the corresponding indices will be concatenated to form the
	        vision features.
	    image_seq_length (`int`, *optional*, defaults to 576):
	        Sequence length of one image embedding.
	    multimodal_projector_bias (`bool`, *optional*, defaults to `True`):
	        Whether to use bias in the multimodal projector.
	"""

	model_type = "llava"
	sub_configs = {"text_config": AutoEasyDeLConfig, "vision_config": AutoEasyDeLConfig}

	def __init__(
		self,
		vision_config=None,
		text_config=None,
		image_token_index=32000,
		projector_hidden_act="gelu",
		vision_feature_select_strategy="default",
		vision_feature_layer=-2,
		image_seq_length=576,
		multimodal_projector_bias=True,
		**kwargs,
	):
		self.image_token_index = image_token_index
		self.projector_hidden_act = projector_hidden_act
		self.image_seq_length = image_seq_length

		if vision_feature_select_strategy not in ["default", "full"]:
			raise ValueError(
				"vision_feature_select_strategy should be one of 'default', 'full'."
				f"Got: {vision_feature_select_strategy}"
			)

		self.vision_feature_select_strategy = vision_feature_select_strategy
		self.vision_feature_layer = vision_feature_layer

		if isinstance(vision_config, dict):
			vision_config["model_type"] = (
				vision_config["model_type"]
				if "model_type" in vision_config
				else "clip_vision_model"
			)
			vision_config = registry.get_config(vision_config["model_type"])(**vision_config)
		elif vision_config is None:
			vision_config = registry.get_config("clip_vision_model")(
				intermediate_size=4096,
				hidden_size=1024,
				patch_size=14,
				image_size=336,
				num_hidden_layers=24,
				num_attention_heads=16,
				vocab_size=32000,
				projection_dim=768,
			)

		self.vision_config = vision_config

		if isinstance(text_config, dict):
			text_config["model_type"] = (
				text_config["model_type"] if "model_type" in text_config else "llama"
			)
			text_config = registry.get_config(text_config["model_type"])(**text_config)
		elif text_config is None:
			text_config = registry.get_config("llama")()

		self.text_config = text_config
		self.multimodal_projector_bias = multimodal_projector_bias

		super().__init__(**kwargs)

	def get_partition_rules(self, *args, **kwargs):
		"""
		Get the partition rules for distributed training by combining the partition rules
		from both the text and vision configurations.

		This method retrieves the partition rules from the text_config and vision_config
		components and combines them to create a comprehensive set of rules for the entire
		multimodal model.

		Args:
		    *args: Variable length argument list to be passed to the text and vision configs.
		    **kwargs: Arbitrary keyword arguments to be passed to the text and vision configs.

		Returns:
		    tuple: A combined tuple of partition rules from both text and vision configurations.
		"""
		tp = self.text_config.get_partition_rules(*args, **kwargs)
		vp = self.vision_config.get_partition_rules(*args, **kwargs)
		return tp + vp
